<?php
/*
	Project: Task Scheduler
	Author : karapuz <support@ka-station.com>

	Version: 3 ($Revision: 139 $)


	This task implements the ability to download files via ftp and http protocols from external
	servers. After downloading the file the script can run the next task which may start
	an import routine.
*/
namespace extension\ka_extensions\tasks;

class ModelDownloadFile extends \KaModel {

	/*
		RETURNS:
			false - on error
			array - on success. It looks like:
				array(
					'status'
						'http_version'  =>
						'status_code'   =>
						'reason_phrase' =>
					'headers'
						'<hdr1>' => value
						'<hdr2>' => value
				)
	*/
	protected function parseHttpHeader($header) {
	
		if (!preg_match("/^(.*)\s(.*)\s(.*)\x0D\x0A/U", $header, $matches)) {
			return false;
		}

		$status = array(
			'http_version'  => $matches[1],
			'status_code'   => $matches[2],
			'reason_phrase' => $matches[3]
		);
		
		$headers = array();		
		$header_lines = explode("\x0D\x0A", $header);
		
		foreach ($header_lines as $line) {
			$pair        = array();
			$value_start = strpos($line, ': ');
			$name        = substr($line, 0, $value_start);
			$value       = substr($line, $value_start + 2);
						
			$headers[$name] = $value;
		}
		
		$result = array(
			'status' => $status,
			'headers' => $headers
		);
		
		return $result;					
	}

	//
	// http://www.w3.org/Protocols/rfc2616/rfc2616-sec6.html
	//
	public function getFileContentsByUrl($url, $params = array()) {

		$this->lastError = '';
		$message = null;
		
		if (function_exists('curl_init')) {

			$tmp_url        = $url;
			$redirect_count = 0;
						
			do {			
				$headers = '';
				$message = null;
				
				$curl = curl_init($tmp_url);

				curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
				curl_setopt($curl, CURLOPT_HEADER, true);
				curl_setopt($curl, CURLOPT_TIMEOUT, 360);
				curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
				curl_setopt($curl, CURLOPT_SSL_VERIFYHOST, false);

				if (!empty($params['webauth_username']) && !empty($params['webauth_password'])) {
					curl_setopt($curl, CURLOPT_USERPWD, $params['webauth_username'] . ":" . $params['webauth_password']);  
				}
				
				$parsed_url = parse_url($url);
				$parsed_tmp_url = parse_url($tmp_url);

				// add custom headers to emulate a regular user activity so it will prevent bans
				// by the user-agent string
				//
				$opt_headers = array();
				if (preg_match("/dropbox\.com/i", $parsed_url['host'])) {
					$opt_headers[] = "User-Agent: Wget/1.11.4";
				} else {
					$opt_headers[] = "User-Agent: Mozilla/5.0 (Windows NT 6.1; WOW64; rv:42.0) Gecko/20100101 Firefox/42.0";
				}
				
				$opt_headers[] = "User-Agent: Mozilla/5.0 (Windows NT 6.1; WOW64; rv:42.0) Gecko/20100101 Firefox/42.0";
 				$opt_headers[] = "Accept: */*";
 				$opt_headers[] = "Host: " . $parsed_tmp_url['host'];
 				$opt_headers[] = "Connection: Keep-Alive";
				if (!empty($opt_headers)) {
					curl_setopt($curl, CURLOPT_HTTPHEADER, $opt_headers);
				}

				$response = curl_exec($curl);
				curl_close($curl);

				if (empty($response)) {
					$this->lastError = "ERROR:" . curl_error($curl) . " (" . curl_errno($curl) . ")";
				}

				$msg_start    = strpos($response, "\x0D\x0A\x0D\x0A");
				$header_block = substr($response, 0, $msg_start);
				$headers      = $this->parseHttpHeader($header_block);				
				if (empty($headers)) {
					break;
				}

				if ($headers['status']['status_code'] >= 200 && $headers['status']['status_code'] < 300) {
					$message = substr($response, $msg_start+4);
					break;
					
				} elseif ($headers['status']['status_code'] >= 300 && $headers['status']['status_code'] < 400) {
					$tmp_url = $headers['headers']['Location'];
					continue;
				} else {
					$this->lastError = 'Invalid status code: ' . $headers['status']['status_code'];
					break;
				}
				
			} while (++$redirect_count < 5);
			
		} else {

			if (ini_get('allow_url_fopen')) {
				$message = file_get_contents($url);
			}
		}

		return $message;
	}


	/*
		returns a list of supported operations.
	*/
	public function requestSchedulerOperations() {
	
		$ret = array(
			'download_ftp'  => 'Download file through FTP',
			'download_http' => 'Download file from HTTP URL'
		);

		return $ret;
	}
	
	
	public function requestSchedulerOperationParams($op) {
	
		$ret = false;
	
		if ($op == 'download_ftp') {
			$ret = array(
				'ftp_url' => array(
					'title' => 'FTP Server URL',
					'type'  => 'text',
					'required' => true,
					'descr' => 'Example: ftp.mysite.com. Do NOT specify any protocol/port here',
				),
				
				'ftp_username' => array(
					'title' => 'FTP Username',
					'type'  => 'text',
					'required' => true,
				),
				'ftp_password' => array(
					'title' => 'FTP Password',
					'type' => 'password',
					'required' => false,
				),
				'server_file' => array(
					'title' => 'Path to file on FTP server',
					'type' => 'text',
					'required' => true,
					'descr' => 'Example: /my_files/data.csv',
				),
			);

			$tasks = $this->model_extension_ka_extensions_ka_scheduler_ka_tasks->getTasks();
			
			$options = array();			
			foreach ($tasks as $task) {
				if (in_array($task['operation'], array('import'))) {
					$options[$task['task_id']] = $task['name'];
				}
			}
			
			$ret['next_task_id']['options'] = $options;
			
		} elseif ($op == 'download_http') {
		
			$ret = array(
				'http_url' => array(
					'title' => 'HTTP URL',
					'type'  => 'text',
					'required' => true,
					'descr' => 'Example: http://www.example.com/downloads/stock.csv',
				),
				'webauth_username' => array(
					'title' => 'WebAuth Username',
					'type'  => 'text',
					'required' => false,
					'descr' => "Web authorization can be provided. it is optional"
				),
				'webauth_password' => array(
					'title' => 'WebAuth  Password',
					'type' => 'password',
					'required' => false,
				),
			);
		}

		$ret['local_file'] = array(
			'title' => 'Relative path to downloaded file (within store directory)',
			'type' => 'text',
			'required' => true,
			'descr' => 'Example: downloaded/file.csv',
		);

		$ret['unpack'] = array(
			'title' => 'Unpack zip file',
			'type' => 'select',
			'required' => false,
			'descr' => 'If the file is a zip archive, it can be unpacked after downloading.',
			'options' => array(
				'0' => 'Do nothing',
				'1' => 'Unpack archive'
			),
		);
		
		$ret['next_task_id'] = array(
			'title'    => 'Call the task on successful completion',
			'type'     => 'select',
			'required' => false,
			'descr'    => 'Import script can be called automatically when the file is downloaded'
		);
		
		$tasks = $this->model_extension_ka_extensions_ka_scheduler_ka_tasks->getTasks();
		
		$options = array();			
		foreach ($tasks as $task) {
			if (in_array($task['operation'], array('import'))) {
				$options[$task['task_id']] = $task['name'];
			}
		}
		
		$ret['next_task_id']['options'] = $options;

		return $ret;
	}

	
	protected function openFtp($url, $username, $password) {

		if (($conn_id = ftp_connect($url)) == false) {
			return false;
		}
		
		// login with username and password
		if (!ftp_login($conn_id, $username, $password)) {
			ftp_close($conn_id);
			return false;
		}
		
		//set passive mode
		// TODO: make it a configurable setting
		//
		ftp_pasv($conn_id, true);

		return $conn_id;
	}

	
	protected function downloadFtpFile($tmp_file, $params, &$stat) {
		
		if (!($conn_id = $this->openFtp($params['ftp_url'], $params['ftp_username'], $params['ftp_password']))) {
			$stat = array(
				'Result' => 'FAILED: Cannot open server with specified access information',
			);
			return false;
		}
	
		// try to download $server_file and save to
		if (!ftp_get($conn_id, $tmp_file, $params['server_file'], FTP_BINARY)) {
			return false;
		}		

		if (!($conn_id = $this->openFtp($params['ftp_url'], $params['ftp_username'], $params['ftp_password']))) {
			$stat = array(
				'Result' => 'FAILED: Cannot open server with specified access information',
			);
			return $res;
		}

		// try to download $server_file and save to $local_file
		if (!ftp_get($conn_id, $tmp_file, $params['server_file'], FTP_BINARY)) {
			$stat['Result'] = 'FAILED: cannot download the file.';
			return false;
			
		} else {
			$stat['Result'] = 'SUCCESS: File is downloaded successfully.';
		}

		ftp_close($conn_id);
				
		return $tmp_file;
	}


	protected function downloadHttpFile($tmp_file, $params, &$stat) {

		$this->lastError = "";
		$contents = $this->getFileContentsByUrl($params['http_url'], $params);

		if (empty($contents)) {
			$stat['Result'] = "File content is empty for url: " . $params['http_url'];
			if (!empty($this->lastError)) {
				$stat['Result'] .= "\n Error: " . $this->lastError;
			}

			return false;
		}

		if (!file_put_contents($tmp_file, $contents)) {
			$stat['Result'] = "Sorry, file cannot be saved on the server as $tmp_file";
			return false;
		}

		return $tmp_file;
	}
			
	/*
		$operation - operation name (string)
		$params    - array of parameters
		$stat      - array of returned strings for displaying results of the operation
		
		RETURNS:
			$return - text code (
					'finished'      - operation is complete
					'not_finished'  - still working (additional calls needed)
					)
			
			$return - array(
					'result'        - text code (finished/not_finished)
					'next_task_id'  - task_id to request on completion
	*/
	public function runSchedulerOperation($operation, $params, &$stat) {

		$result = array(
			'result' => 'not_finished',
		);
		
		$this->lastError = '';		
		if (empty($this->session->data['stage'])) {
			$this->session->data['stage'] = 'download';
			$stat = array();			
		}

		if ($this->session->data['stage'] == 'download') {
			$tmp_file = tempnam(DIR_CACHE, 'kadnl_');
			if ($operation == 'download_ftp') {
				$ret = $this->downloadFtpFile($tmp_file, $params, $stat);
				
			} elseif ($operation == 'download_http') {
				$ret = $this->downloadHttpFile($tmp_file, $params, $stat);
			}

			if ($ret == false) {
				$result['result'] = 'finished';
				return $result;
			}
			
			if (file_exists($params['local_file'])) {
				unlink($params['local_file']);
			}
			
			$target_file = dirname(DIR_SYSTEM) . '/' . $params['local_file'];
			if (!rename($ret, $target_file)) {
				$stat['Result'] = 'Renaming file failed';
				$result['result'] = 'finished';
				return $result;
			}
			
			$this->session->data['target_file'] = $target_file;

			$stat = array(
				'File size' => filesize($target_file),
				'Result' => 'SUCCESS: File is downloaded successfully.',			
			);
			
			// go to a next step if we need to unpack the file
			if (!empty($params['unpack'])) {
				$this->session->data['stage'] = 'unpack';
				return $result;
			}
			
		} elseif ($this->session->data['stage'] == 'unpack') {
			if (!$this->unpackFile($this->session->data['target_file'])) {
				$stat['Error'] = 'Unpack failed.' . $this->getLastError();
			}
		}

		$result['result'] = 'finished';
		
		if (!empty($params['next_task_id'])) {
			$result['next_task_id'] = $params['next_task_id'];
		}
		
		return $result;
	}
	
	
	protected function unpackFile($file) {
		if (!class_exists('\ZipArchive')) {
			$this->lastError = 'ZipArchive class does not exist in your PHP installation. Please adjust server configuration.';
			return false;
		}
		
		$zip = new \ZipArchive;
		$res = $zip->open($file);
		if ($res === TRUE) {
		    $zip->extractTo(dirname($file));
		    $zip->close();		    
		    return true;
		}	
		
		$this->lastError = "ZipArchive failed with an error ($res)";
		return false;
	}
}

class_alias(__NAMESPACE__ . '\ModelDownloadFile', 'ModelExtensionKaExtensionsTasksDownloadFile');